import { useRef, useState, useCallback, useMemo, useEffect } from "react";
import { useDataLoaderContext } from "./DataLoaderProvider.js";
import { StatusEnum } from "./constants.js";
import { marshalQueryKey } from "./helpers.js";
const useInfiniteDataLoader = (baseKey, method, baseParams, pageParamKey, config) => {
  const {
    getOrAddRequest,
    computeKey,
    onError: onGlobalError,
    defaultDatalifetime
  } = useDataLoaderContext();
  const {
    enabled = true,
    onError,
    onSuccess,
    keepPreviousData = false,
    initialData,
    dataLifetime,
    getNextPage
  } = config ?? {};
  const computedDatalifetime = dataLifetime ?? defaultDatalifetime;
  const requestRefs = useRef([]);
  const [page, setPage] = useState(baseParams[pageParamKey]);
  const nextPageRef = useRef(page);
  const getNextPageFnRef = useRef(
    (...params) => getNextPage ? getNextPage(...params) : void 0
  );
  const getParamsRef = useRef(() => ({
    ...baseParams,
    [pageParamKey]: nextPageRef.current
  }));
  const getMethodRef = useRef(() => method(getParamsRef.current()));
  const getOnSuccessRef = useRef(
    (...params) => onSuccess?.(...params)
  );
  const getOnErrorRef = useRef(
    (err) => onError?.(err) ?? onGlobalError?.(err)
  );
  const [, setCounter] = useState(0);
  const forceRerender = useCallback(() => {
    setCounter((current) => current + 1);
  }, []);
  const baseQueryKey = useMemo(() => marshalQueryKey(baseKey), [baseKey]);
  useEffect(
    () => () => {
      requestRefs.current.forEach(
        (request2) => request2.removeObserver(forceRerender)
      );
    },
    [forceRerender]
  );
  const getCurrentRequest = () => {
    const currentQueryKey = marshalQueryKey([
      baseQueryKey,
      "infinite",
      page
    ]);
    requestRefs.current.forEach(
      (request2) => !request2.key.startsWith(computeKey(baseQueryKey)) ? request2.removeObserver(forceRerender) : void 0
    );
    requestRefs.current = requestRefs.current.filter(
      ({ key }) => key.startsWith(computeKey(baseQueryKey))
    );
    const requestInRef = requestRefs.current.find(
      (request2) => request2.key.endsWith(currentQueryKey)
    );
    if (!requestInRef) {
      const request2 = getOrAddRequest(currentQueryKey, {
        enabled,
        method: getMethodRef.current
      });
      requestRefs.current.push(request2);
      request2.addObserver(forceRerender);
      return request2;
    }
    return requestInRef;
  };
  const request = getCurrentRequest();
  const needLoad = useMemo(
    () => !!(enabled && (!request.dataUpdatedAt || !computedDatalifetime || request.dataUpdatedAt && computedDatalifetime && request.dataUpdatedAt + computedDatalifetime < Date.now())),
    [enabled, request.dataUpdatedAt, computedDatalifetime]
  );
  const optimisticIsLoadingRef = useRef(needLoad);
  const previousDataRef = useRef(request.data);
  const isLoading = requestRefs.current.some(
    (req) => req.status === StatusEnum.LOADING || optimisticIsLoadingRef.current
  );
  const isLoadingFirstPage = (request.status === StatusEnum.LOADING || optimisticIsLoadingRef.current) && requestRefs.current.length <= 1;
  const isSuccess = request.status === StatusEnum.SUCCESS;
  const isError = request.status === StatusEnum.ERROR;
  const isIdle = requestRefs.current.every(
    (req) => req.status === StatusEnum.IDLE && !enabled
  );
  const reload = useCallback(async () => {
    await Promise.all(
      requestRefs.current.map(
        (req) => req.load(true).then(getOnSuccessRef.current).catch(getOnErrorRef.current)
      )
    );
  }, []);
  const loadMore = useCallback(() => {
    const nextPage = nextPageRef.current;
    if (nextPage) {
      setPage(() => nextPage);
      getParamsRef.current = () => ({
        ...baseParams,
        [pageParamKey]: nextPage
      });
    }
  }, [baseParams, pageParamKey]);
  useEffect(() => {
    request.method = () => method(getParamsRef.current());
  }, [method, request]);
  useEffect(() => {
    if (keepPreviousData) {
      previousDataRef.current = request.data;
    }
  }, [request.data, keepPreviousData]);
  useEffect(() => {
    setPage(() => baseParams[pageParamKey]);
    nextPageRef.current = baseParams[pageParamKey];
    getParamsRef.current = () => ({
      ...baseParams,
      [pageParamKey]: nextPageRef.current
    });
  }, [baseParams, pageParamKey]);
  useEffect(() => {
    if (needLoad) {
      const onSuccessLoad = getOnSuccessRef.current;
      const onFailedLoad = getOnErrorRef.current;
      request.load().then(async (result) => {
        nextPageRef.current = getNextPageFnRef.current(
          result,
          getParamsRef.current()
        );
        await onSuccessLoad(result);
      }).catch(onFailedLoad);
    }
    optimisticIsLoadingRef.current = false;
  }, [needLoad, request]);
  useEffect(() => {
    getParamsRef.current = () => ({
      ...baseParams,
      [pageParamKey]: nextPageRef.current
    });
  }, [baseParams, pageParamKey]);
  useEffect(() => {
    getOnSuccessRef.current = (...params) => onSuccess?.(...params);
  }, [onSuccess]);
  useEffect(() => {
    getOnErrorRef.current = (err) => onError?.(err) ?? onGlobalError?.(err);
  }, [onError, onGlobalError]);
  useEffect(() => {
    getNextPageFnRef.current = (...params) => getNextPage ? getNextPage(...params) : void 0;
  }, [getNextPage]);
  const data = useMemo(
    () => ({
      isIdle,
      isError,
      isLoading,
      isSuccess,
      hasNextPage: nextPageRef.current !== void 0,
      isLoadingFirstPage,
      data: isLoadingFirstPage || [...requestRefs.current].filter((dataloader) => !!dataloader.data).length === 0 ? initialData : [...requestRefs.current].filter((dataloader) => !!dataloader.data).map((dataloader) => dataloader.data),
      error: request.error,
      reload,
      loadMore
    }),
    [
      initialData,
      isIdle,
      isLoading,
      isSuccess,
      isError,
      request.error,
      isLoadingFirstPage,
      reload,
      loadMore
    ]
  );
  return data;
};
export {
  useInfiniteDataLoader
};
